"""
Pynexafs.nexafs.scan: A set of classes for handling synchrotron scan data.

This module defines validated scan objects that can perform intelligent operations
on synchrotron data. These objects are constructed from a `parser` object that contains
the raw data and metadata.

See Also
--------
pynexafs.parsers.parserBase : Base class for synchrotron data parsers.
"""

# Stdlib imports
from __future__ import annotations  # For type hinting within class definitions
import abc
import datetime
import os
import io
from typing import TYPE_CHECKING, Self, override, overload

# External imports
import numpy.typing as npt
import numpy as np

try:
    import matplotlib.figure
    import matplotlib.axes
    import matplotlib.pyplot as plt

    HAS_MPL = True
except ImportError:
    HAS_MPL = False

# Internal imports
from pyNexafs.types import dtype

if TYPE_CHECKING:
    from ..parsers import parserBase


class scanAbstract(metaclass=abc.ABCMeta):
    """
    Abstract class for defining properties of a scan object.

    Base abstract class includes x,y attributes for data, errors, labels and units.
    """

    def __init__(self) -> None:
        # Internal variables
        self._x: npt.NDArray | None = None
        self._y: npt.NDArray | None = None
        self._x_errs: npt.NDArray | None = None
        self._y_errs: npt.NDArray | None = None
        self._x_label: str | None = None
        self._x_unit: str | None = None
        self._y_labels: list[str | None] | None = None
        self._y_units: list[str | None] | None = None
        return

    @abc.abstractmethod
    def copy(self) -> Self:
        """
        Create a copy of the scan object.

        Implemented method should call this method to copy the data attributes, and implement
        copying of any subclass specific attributes.

        Returns
        -------
        Self
            A copy of the scan object with unique data.
        """
        newobj = self.__class__()

        # Copy Data
        newobj._x = self._x.copy() if self._x is not None else None
        newobj._y = self._y.copy() if self._y is not None else None
        newobj._y_errs = self._y_errs.copy() if self._y_errs is not None else None
        newobj._x_errs = self._x_errs.copy() if self._x_errs is not None else None

        # Copy Labels and Units
        newobj._x_label = self._x_label
        newobj._x_unit = self._x_unit
        if self._y_labels is not None:
            if isinstance(self._y_labels, list):
                newobj._y_labels = [label for label in self._y_labels]
            else:  # string
                newobj._y_labels = self._y_labels
        else:
            newobj._y_labels = None
        if self._y_units is not None:
            if isinstance(self._y_units, list):
                newobj._y_units = [unit for unit in self._y_units]
            else:  # string
                newobj._y_units = self._y_units
        else:
            newobj._y_units = None
        return newobj

    @property
    def ctime(self) -> datetime.datetime | None:
        """
        The creation time of the file as a datetime object.

        Returns
        -------
        datetime.datetime | None
            The creation time of the file.
            None if not defined.

        Raises
        ------
        NotImplementedError
            If the ctime property is not implemented in the subclass.
        """
        raise NotImplementedError("ctime property not implemented.")

    @property
    def mtime(self) -> datetime.datetime | None:
        """
        The modification time of the file as a datetime object.

        Returns
        -------
        datetime.datetime | None
            The modification time of the file.
            None if not defined.

        Raises
        ------
        NotImplementedError
            If the mtime property is not implemented in the subclass.
        """
        raise NotImplementedError("mtime property not implemented.")

    @property
    def x(self) -> npt.NDArray | None:
        """
        Property for the 1D X data (beam energy) of the scan.

        Parameters
        ----------
        val : npt.NDArray | list[int|float]
            A new 1D array to define beam energies. Removes x_errs if new dimensions don't match.

        Returns
        -------
        npt.NDArray | None
            The beam energies of the scan.
        """
        return self._x

    @x.setter
    def x(self, val: npt.NDArray | list[int | float]) -> None:
        if isinstance(val, np.ndarray):
            self._x = val.copy()
        elif isinstance(val, list):
            self._x = np.array(val)
        else:
            raise ValueError("Is not a list or a np.ndarray.")
        # Remove errors if dimensions have changed.
        if self._x_errs is not None and self._x_errs.shape != self._x.shape:
            self._x_errs = None

    @property
    def y(self) -> npt.NDArray | None:
        """
        Property for the Y data of the scan.

        Parameters
        ----------
        val : npt.NDArray | list[int|float] | None
            A numpy array of data or a list of datapoints.

        Returns
        -------
        npt.NDArray
            The beam energies of the scan.
        """
        return self._y

    @y.setter
    def y(
        self, vals: npt.NDArray | list[list[int | float]] | list[int | float] | None
    ) -> None:
        # Set Y values
        if isinstance(vals, np.ndarray):
            if len(vals.shape) == 2:
                self._y = vals.copy()
            else:
                # Add additional axis to make y 2D instead of 1D
                self._y = vals.copy()[:, np.newaxis]
        elif isinstance(vals, list):
            if isinstance(vals[0], list):
                self._y = np.array(vals)
            else:
                # Assuming int/floats, add additional axis as well.
                self._y = np.array(vals)[:, np.newaxis]
        elif vals is None:
            self._y = None
            self._y_errs = None
            self._y_labels = None
            return
        else:
            raise ValueError("Is not a list or a np.ndarray.")

        # Remove errs if number of yvals have changed.
        if self.y_errs is not None and self.y_errs.shape != self._y.shape:
            self.y_errs = None

        # Remove unit/labels if number of yvals have changed
        ylabels = self.y_labels
        if ylabels is not None and len(ylabels) != self._y.shape[1]:
            self.y_labels = None
        yunits = self.y_units
        if yunits is not None and len(yunits) != self._y.shape[1]:
            self.y_units = None

    @property
    def y_errs(self) -> npt.NDArray | None:
        """
        Property for error data corresponding to the Y values.

        Parameters
        ----------
        errs : np.NDArray | None
            Error values in a numpy array matching y shape, or None to remove errors.

        Returns
        -------
        np.NDArray | None
            Errors corresponding to y values or None if not defined.
        """
        return self._y_errs

    @y_errs.setter
    def y_errs(
        self, errs: npt.NDArray | list[list[int | float]] | list[int | float] | None
    ) -> None:
        if errs is None:
            self._y_errs = None
            return
        y = self.y

        if isinstance(errs, np.ndarray):
            if len(errs.shape) == 2:
                errs = errs.copy()
            else:
                # Add additional axis to make y 2D instead of 1D
                errs = errs.copy()[:, np.newaxis]
        elif isinstance(errs, list):
            if isinstance(errs[0], list):
                errs = np.array(errs)
            else:
                # Assuming int/floats, add additional axis as well.
                errs = np.array(errs)[:, np.newaxis]
        else:
            raise ValueError("Is not a list of numbers or a np.ndarray.")

        if y is not None:
            if y.shape == errs.shape:
                self._y_errs = errs
            else:
                raise ValueError(
                    f"Shape of errors {errs.shape} doesn't match existing y shape {y.shape}."
                )
        else:
            self._y_errs = errs  # Allow setting errs if y is None.
        return

    @property
    def x_errs(self) -> npt.NDArray | None:
        """
        Property for error data corresponding to the X values.

        Parameters
        ----------
        errs : np.NDArray | None
            Error values in a numpy array matching x shape, or None to remove errors.

        Returns
        -------
        np.NDArray | None
            Errors corresponding to x values or None if not defined.
        """
        return self._x_errs

    @x_errs.setter
    def x_errs(self, errs: npt.NDArray | None) -> None:
        if errs is None:
            self._x_errs = None
        x = self.x
        if isinstance(errs, np.ndarray):
            if x is not None:
                if errs.shape == x.shape:
                    self._x_errs = errs.copy()
                else:
                    raise ValueError(
                        f"Shape of errors {errs.shape} doesn't match existing x shape {x.shape}."
                    )
            else:
                self._x_errs = errs.copy()  # Allow setting errs if x is None.
        else:
            raise ValueError("Is not a np.ndarray.")

    @property
    def x_label(self) -> str:
        """
        Property for the label of the x axis.

        Parameters
        ----------
        label : str | None
            String to replace existing x label.

        Returns
        -------
        str
            The x label.
            Defaults to "Energy" if no label is defined.
        """
        if self._x_label is None:
            return "Energy"
        return self._x_label

    @x_label.setter
    def x_label(self, label: str | None) -> None:
        if isinstance(label, str):
            self._x_label = label
        elif label is None:
            self._x_label = None
        else:
            raise ValueError(f"New label `{label}` is not string.")

    @property
    def y_labels(self) -> list[str | None] | None:
        """
        Property for the labels of the y axis.

        In the event new y data is supplied and doesn't match number of existing y_labels,
        generic y_labels are created in the form 'Data Col. N' where N is the column index.

        Parameters
        ----------
        labels : list[str | None] | None
            String to replace existing y labels. Must match number of columns in the y attribute.

        Returns
        -------
        list[str | None] | None
            A direct reference to the y labels list. If y data has been redefined
            without redefining y_labels, y labels are generated in the form 'Data Col. N'.
        """
        if self._y_labels is not None:
            return self._y_labels
        else:
            y = self.y
            if y is None:
                return None

            chars_col_len = int(
                np.ceil(np.log10(y.shape[1] + 1))
            )  # Gets number of characters for a given number.
            self._y_labels = [
                "Data Col. " + str(i + 1).zfill(chars_col_len)
                for i in range(y.shape[1])
            ]
            return self._y_labels

    @y_labels.setter
    def y_labels(self, labels: list[str | None] | None) -> None:
        if isinstance(labels, list) and np.all(
            [isinstance(label, str) or label is None for label in labels]
        ):
            y = self.y
            if y is not None:
                if len(labels) == y.shape[1]:
                    self._y_labels = labels.copy()
                else:
                    raise ValueError(
                        f"Number of labels ({len(labels)}) does not match y data columns ({y.shape[1]})."
                    )
            else:
                self._y_labels = labels.copy()  # Allow setting labels if y is None.
        elif labels is None:
            self._y_labels = None
        else:
            raise ValueError(
                f"Provided `labels` {labels} is not a list of only strings."
            )
        return

    @property
    def x_unit(self) -> str:
        """
        Property for the unit of the x data.

        Parameters
        ----------
        unit : str
            New string for the x data unit.

        Returns
        -------
        str
            The label of the x data. Defaults to "eV" if no unit is defined.
        """
        if self._x_unit is None:
            return "eV"
        return self._x_unit

    @x_unit.setter
    def x_unit(self, unit: str | None) -> None:
        if isinstance(unit, str):
            self._x_unit = unit
        elif unit is None:
            self._x_unit = None
        else:
            raise ValueError(f"Provided `unit` {unit} is not a string.")
        return

    @property
    def y_units(self) -> list[str | None] | None:
        """
        Property for the units of the y data.

        Parameters
        ----------
        units : list[str | None] | None
            A list of strings to replace existing y units. Must match number of columns in the y attribute.

        Returns
        -------
        list[str | None] | None
            A direct reference to the y units list. If y data has been redefined
            without redefining y_units, y units are generated in the form 'Data Col. N'.
        """
        return self._y_units

    @y_units.setter
    def y_units(self, units: list[str | None] | None):
        if units is None:
            self._y_units = None
        elif isinstance(units, list) and np.all(
            [isinstance(unit, str | None) for unit in units]
        ):
            y = self.y
            if y is not None:
                if len(units) == y.shape[1]:
                    self._y_units = units.copy()
                else:
                    raise ValueError(
                        f"Shape of 'units' ({len(units)}) does not match shape of y data ({y.shape[1]})."
                    )
            else:
                self._y_units = units.copy()  # Allow setting units if y is None.
        else:
            raise ValueError(f"Provided 'units' {units} is not a list of strings.")
        return

    def snapshot(self, columns: int | None = None) -> "matplotlib.figure.Figure":
        """
        Generate a grid of plots, showing all scan data.

        Parameters
        ----------
        columns : int, optional
            Number of columns used to grid data, by default None
            will use the square root of the number of Y channels.

        Returns
        -------
        matplotlib.figure.Figure
            A figure object containing the grid of plots.

        Raises
        ------
        ImportError
            If matplotlib is not installed.
        ValueError
            If Y data is not defined.
        """
        if not HAS_MPL:
            raise ImportError(
                "Matplotlib is not installed, cannot generate a snapshot graphic."
            )

        y = self.y
        if y is None:
            raise ValueError("Y data is not defined.")
        if columns is None:
            columns = int(np.ceil(np.sqrt(y.shape[1])))
        rows = int(np.ceil(y.shape[1] / columns))
        fig, ax = plt.subplots(rows, columns, sharex=True, figsize=(10, 10), dpi=300)  # type: ignore - Covered by HAS_MPL check.

        # Ensure axis array is 2D.
        if not isinstance(ax, np.ndarray):
            ax = np.array([[ax]])
        elif ax.ndim == 1:
            ax = np.array([ax])

        # Plot the data.
        ylabels = self.y_labels
        yunits = self.y_units
        for i, ydata in enumerate(y.T):
            r, c = divmod(i, columns)
            ax[r][c].plot(self.x, ydata)
            if ylabels is not None:
                ax[r][c].set_title(
                    ylabels[i] if yunits is None else f"{ylabels[i]} ({yunits[i]})"
                )
        for i in range(columns):
            ax[rows - 1][i].set_xlabel(self.x_label)
        return fig

    def to_csv(
        self,
        filename: str | None = None,
        delim: str = ",",
        combine_label_unit: bool = True,
    ) -> None | str:
        r"""
        Save the processed data as a CSV (comma-separated values) file.

        TODO: Currently doesn't save errors.

        Parameters
        ----------
        filename : str | None
            The name of the file to save the data to.
            If None, returns the CSV buffer as a string.
        delim : str, optional
            The delimiter to use in the csv file, by default ",".
            Another common option is to use "\t" for tab delimited.
        combine_label_unit : bool, optional
            If True, combines the label and unit in the header, by default True.
            Otherwise label and unit are separate rows.

        Returns
        -------
        None | str
            If filename is `None`, returns the CSV buffer as a string.

        Raises
        ------
        IOError
            If the specified file already exists.
        """
        x = self.x
        if x is None:
            raise ValueError("X data is not defined.")
        y = self.y
        if y is None:
            raise ValueError("Y data is not defined.")

        if filename is None:
            csv_file = io.StringIO()
        else:
            if os.path.exists(filename):
                raise IOError(f"The following filepath already exists:\n{filename}")
            csv_file = open(filename, "w")
        # Write the header
        y_labels = self.y_labels
        y_units = self.y_units
        if combine_label_unit:
            if y_labels is not None and y_units is not None:
                combo = []
                for label, unit in zip(y_labels, y_units):
                    if label is not None and unit is not None:
                        combo.append(f"{label} ({unit})")
                    elif label is not None:
                        combo.append(label)
                    elif unit is not None:
                        combo.append(f"({unit})")
                    else:
                        combo.append("")
                csv_file.write(
                    f"{self.x_label} ({self.x_unit}){delim}{delim.join(combo)}\n"
                )
            elif y_labels is not None:
                combo = [label if label is not None else "" for label in y_labels]
                csv_file.write(
                    f"{self.x_label} ({self.x_unit}){delim}{delim.join(combo)}\n"
                )
            else:
                csv_file.write(f"{self.x_label}{delim}Y Data\n")
        else:
            # Write labels, then units.
            if y_labels is not None:
                combo = [label if label is not None else "" for label in y_labels]
                csv_file.write(f"{self.x_label}{delim}{delim.join(combo)}\n")
            if y_units is not None:
                combo = [unit if unit is not None else "" for unit in y_units]
                csv_file.write(f"{self.x_unit}{delim}{delim.join(combo)}\n")
        # Write the data
        for i in range(len(x)):
            csv_file.write(f"{x[i]}{delim}{delim.join(str(y) for y in y[i])}\n")

        if filename is None:
            csv_file.seek(0)  # Reset the buffer to the beginning
            return csv_file.read()
        else:
            csv_file.close()

    @abc.abstractmethod
    def reload_labels_from_parser(self) -> None:
        """
        Re-load the labels and units from the parser object.

        This method is abstract and must be implemented in subclasses.
        Useful when the user wants to switch from the raw parameter names to useful names.
        Alternatively scan labels can be manually set.
        """
        pass


class scanSimple(scanAbstract):
    """
    Basic interface class for raw data that is not bundled in a parser object.

    This class is used for simple scans where data is provided directly as x and y arrays.

    Parameters
    ----------
    x : npt.NDArray | None
        1D array of x data (e.g., beam energy).
    y : npt.NDArray | None
        2D array of y data (e.g., multiple Y channels). First index is data points, second index is channels.
    *args, **kwargs
        Additional arguments and keyword arguments to pass to the parent class.
    """

    def __init__(
        self, x: npt.NDArray | None, y: npt.NDArray | None, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        x = np.array(x) if x is not None else None  # make unique
        y = np.array(y) if y is not None else None
        # Validate x and y data
        if x is not None:
            if not isinstance(x, np.ndarray) or len(x.shape) != 1:
                raise ValueError("x must be a 1D numpy array.")
        if y is not None:
            if not isinstance(y, np.ndarray) or len(y.shape) > 2:
                raise ValueError("y must be a 1D or 2D numpy array.")
            if len(y.shape) == 1:
                y = y[:, np.newaxis]
        self._x = x
        self._y = y
        return

    def reload_labels_from_parser(self) -> None:
        """
        Implement the abstract method from scanAbstract.

        This method is not applicable for scanSimple as it does not use a parser object.
        It is included to satisfy the abstract base class interface.
        """
        return None

    @override
    def copy(self) -> scanSimple:
        """
        Create a copy of the scan object.

        Data is unique for a `scan_simple` object, so no need to reload parser data unlike scan_abstract.

        Returns
        -------
        scanSimple
            A new instance of scanSimple with copied data.
        """
        x = self.x
        y = self.y
        x = x.copy() if x is not None else None
        y = y.copy() if y is not None else None
        new_obj = scanSimple(x=x, y=y)
        return new_obj


class scanBase(scanAbstract):
    """
    Base class for synchrotron measurements that scans across photon beam energies (eV).

    Links to a `parser` instance that contains `parser.data` and `parser.params`.
    The `parser.COLUMN_ASSIGNMENTS` property must reflect the correct column assignments to x/y data.
    This allows for multiple Y channels, reflecting various collectors that can be used in the beamline for absorption data.

    Parameters
    ----------
    parser : Type[parserBase] | parserBase | None
        A parser object that contains the raw data and metadata.
        Parser data is loaded, and can be re-loaded.
        If None, the scan object is empty and must be loaded from a parser object.
    load_all_columns : bool, optional
        If True, load all columns from the parser object.
        If False, only loads the columns defined in `parser.COLUMN_ASSIGNMENTS`.
        Default is False, as this is more memory efficient and typically only the
        x/y columns are needed for analysis.

    See Also
    --------
    parserBase : Base class for synchrotron data parsers.
    """

    @override
    def __init__(
        self,
        parser: parserBase | None,
        load_all_columns: bool = False,
    ) -> None:
        # Initialise data arrays, including x, y, etc.
        super().__init__()

        # Store parser reference.
        self._parser = parser
        self._parser_class = parser.__class__ if parser is not None else None
        self._all_columns_loaded = (
            False  # internal tracking on whether all columns have been loaded or not.
        )
        self._filepath = parser.filepath if parser is not None else None
        self._mtime = parser.mtime if parser is not None else None
        self._ctime = parser.ctime if parser is not None else None

        # Load data from parser
        if parser is not None:
            parser.to_scan(load_all_columns=load_all_columns, scan_obj=self)
            self._all_columns_loaded = load_all_columns

    def detach_parser(self) -> None:
        """
        Remove the reference to the parser, to reduce memory usage.

        Allows for garbage collector to remove the parser object.
        Especially useful if parsed a large file that has been reduced.
        """
        del self.parser

    @property
    def parser_class(self) -> type[parserBase]:
        """
        Property for the class of the parser object linked to the scan object.

        Returns
        -------
        type[parserBase]
            The class of the parser object linked to the scan object.

        Raises
        ------
        ValueError
            If no parser class has been linked to the scan object.
        """
        if self._parser_class is None:
            raise ValueError("No parser class has been linked to this scan object.")
        return self._parser_class

    @override
    def copy(self, *args, **kwargs) -> Self:
        """
        Create a copy of the scan object.

        Parameters
        ----------
        *args, **kwargs
            Additional arguments and keyword arguments to pass to the constructor of the subclass.

        Returns
        -------
        Self
            A new instance of scanBase with copied data.
        """
        # Create a new instance that copies the data
        newobj = self.__class__(parser=None, *args, **kwargs)
        # Copy the base scan attributes:
        newobj._parser = self._parser  # Keep the same parser reference.
        newobj._all_columns_loaded = self._all_columns_loaded
        # Return the copy
        return newobj

    @property
    def ctime(self) -> datetime.datetime | None:
        """
        Return the creation time of the file as a datetime object.

        Returns
        -------
        datetime.datetime | None
            The creation time of the file.
            None if not defined.
        """
        return self._ctime

    @property
    def mtime(self) -> datetime.datetime | None:
        """
        The modification time of the file as a datetime object.

        Returns
        -------
        datetime.datetime | None
            The modification time of the file.
            None if not defined.
        """
        return self._mtime

    @property
    def filename(self) -> str | None:
        """
        Property for the filename of the scan object.

        Returns
        -------
        str | None
            The filename of the scan object.
            None if not defined.
        """
        return os.path.basename(self._filepath) if self._filepath is not None else None

    @property
    def filepath(self) -> str | None:
        """
        Property for the full filepath of the scan object.

        Returns
        -------
        str | None
            The full filepath of the scan object.
            None if not defined.

        Raises
        ------
        ValueError
            If the filepath is not defined.
        """
        return self._filepath

    @property
    def parser(self) -> parserBase | None:
        """
        Property for the parser object linked to the scan object.

        Parameters
        ----------
        parser : parserBase | None
            A parser object that contains the raw data and metadata.
            Parser data is loaded, and can be re-loaded.
            If None, the scan object is empty and must be loaded from a parser object.

        Returns
        -------
         | parserBase
            The parser object linked to the scan object.
        """
        return self._parser

    @parser.setter
    def parser(self, parser: parserBase | None) -> None:
        self._parser = parser
        if parser is not None:
            self._filename = parser.filename
            self._filepath = parser.filepath
            self._mtime = parser.mtime
            self._ctime = parser.ctime
            self._parser_class = parser.__class__
        return

    @parser.deleter
    def parser(self) -> None:
        self._parser = None
        return

    # TODO Access parser class without attached object.
    # @property
    # def parser_class

    def reload(self, load_all_columns: bool | None = None) -> None:
        """
        Reload all data from the parser object.

        If `load_all_columns` is provided, it will override the current attribute setting `_all_columns_loaded`
        otherwise uses the `_all_columns_loaded` attribute to determine to load all columns or only those
        defined in `COLUMN_ASSIGNMENTS`, which is defaulted to `false`.
        Also reloads the x, y, y_errs, x_errs, labels and units from the parser object.

        Parameters
        ----------
        load_all_columns : bool, optional
            If True/False, updates the `_all_columns_loaded` attribute.
            If True, load all columns from the parser object.
            If False, only loads the columns defined in `parser.COLUMN_ASSIGNMENTS`.
            If None, uses the current value of `_all_columns_loaded`.

        Raises
        ------
        ValueError
            If the parser object is not defined, cannot reload data.
        """
        parser = self.parser
        if parser is None:
            raise ValueError("Parser object is not defined, cannot reload data.")
        # Run the parser-scan conversion method, applying to self.
        parser.to_scan(
            load_all_columns=self._all_columns_loaded
            if load_all_columns is None
            else load_all_columns,
            scan_obj=self,
        )

    def reload_labels_from_parser(self) -> None:
        """
        Re-load the labels and units from the parser object.

        Useful when the user wants to switch from the raw parameter names to useful names.
        Alternatively scan labels can be manually set.
        """
        parser = self.parser
        if parser is None:
            raise ValueError("Parser object is not defined, cannot reload data.")
        parser.to_scan(
            load_all_columns=self._all_columns_loaded, scan_obj=None, only_labels=True
        )
        return

    @property
    def channel_map(self) -> dict[str, dtype]:
        """
        A list of the available `dtype` channels available on the object.

        Uses the pre-defined `CHANNEL_MAP` from the parser class to find relevant channels.

        Returns
        -------
        dict[str, dtype]
            A list of the enumerate datatypes.
        """
        parser = self.parser
        if parser is None:
            raise ValueError("Parser object is not defined, cannot get channel map.")
        return parser.CHANNEL_MAP

    @overload
    def _channels_indexes(
        self,
        relabel: bool | None,
        reduced_labels: list[str | None] | None,
        specific: None = None,
    ) -> dict[dtype, int]: ...

    @overload
    def _channels_indexes(
        self,
        relabel: bool | None,
        reduced_labels: list[str | None] | None,
        specific: dtype,
    ) -> int: ...

    def _channels_indexes(
        self,
        relabel: bool | None = None,
        reduced_labels: list[str | None] | None = None,
        specific: dtype | None = None,
    ) -> dict[dtype, int] | int:
        """
        The mapping of the available dtype channels.

        If a specific dtype is requested, the index of that dtype channel is returned instead of the full mapping.

        Parameters
        ----------
        relabel : bool | None
            Whether to check relabelled names or not.
            By default None, which defers to using the default class setting.
        reduced_labels : list[str | None] | None
            Replaces use of default `parser.labels`, particularly useful when using reduced data.
        specific : dtype, optional
            A specific dtype to search for if provided.

        Returns
        -------
        dict[dtype, int]
            The mapping of the available dtype channels.

        Raises
        ------
        ValueError
            If the parser class is not defined, cannot get channel mapping information.
        ValueError
            If a specific dtype is requested but not found in the labels.
        """
        pclass = self._parser_class
        if pclass is None:
            raise ValueError(
                "Parser class is not defined, cannot get channel mapping information."
            )

        parser = self.parser
        if relabel is None:
            if parser is not None:
                relabel = parser.relabel
            elif pclass is not None:
                relabel = pclass.relabel
            else:
                # No relabelling available.
                relabel = False

        specific_search = specific is not None
        mapping: dict[dtype, int] = {}

        labels = self.y_labels if reduced_labels is None else None
        if not specific_search:
            if labels is not None:
                for dt, label in pclass._CHANNEL_MAP_REVERSE.items():
                    # Name directly in channel labels
                    if labels is not None and label in labels:
                        mapping[dt] = labels.index(label)
                    elif relabel:
                        # Check the RELABELS of the channel map label.
                        if label in pclass.RELABELS:
                            rlabel = pclass.RELABELS[label]
                            # Check the relabel
                            if rlabel in labels:
                                mapping[dt] = labels.index(rlabel)
                            else:
                                # Check each
                                reverse = pclass.RELABELS_REVERSE[rlabel]
                                if isinstance(reverse, tuple):
                                    for olabel in reverse:
                                        if olabel in labels:
                                            mapping[dt] = labels.index(olabel)
        else:
            if labels is not None:
                label = pclass._CHANNEL_MAP_REVERSE.get(specific, None)
                if label is not None and label in labels:
                    return labels.index(label)
                elif relabel and label is not None:
                    if label in pclass.RELABELS:
                        rlabel = pclass.RELABELS[label]
                        if rlabel in labels:
                            return labels.index(rlabel)
                        else:
                            reverse = pclass.RELABELS_REVERSE[rlabel]
                            if isinstance(reverse, tuple):
                                for olabel in reverse:
                                    if olabel in labels:
                                        return labels.index(olabel)
            raise ValueError(f"Channel {specific} not found in labels.")
        return mapping

    @property
    def channels(self, relabel: bool | None = None) -> list[dtype]:
        """
        A list of the available `dtype` channels available on the object.

        Uses the pre-defined `CHANNEL_MAP` to find relevant channels.

        Parameters
        ----------
        relabel : bool | None
            Whether to check relabelled names or not.
            By default None, which defers to using the default class setting.

        Returns
        -------
        list[dtype]
            A list of the enumerate datatypes.
        """
        if relabel is None:
            relabel = self.__class__.relabel

        found = []
        labels = self.labels
        if labels is not None:
            for dt, label in self.__class__._CHANNEL_MAP_REVERSE.items():
                # Name directly in channel labels
                if labels is not None and label in labels:
                    found.append(dt)
                elif relabel:
                    # Check the RELABELS of the channel map label.
                    if label in self.RELABELS:
                        rlabel = self.RELABELS[label]
                        # Check the relabel
                        if rlabel in labels:
                            found.append(dt)
                        else:
                            # Check each
                            reverse = self.RELABELS_REVERSE[rlabel]
                            if isinstance(reverse, tuple):
                                for olabel in reverse:
                                    if olabel in labels:
                                        found.append(dt)
        return found

    @property
    def AEY(self) -> npt.NDArray | None:
        """
        The Auger Electron Yield (AEY) channel data if available.

        Returns
        -------
        npt.NDArray | None
            The AEY channel data if available. Otherwise None.
        """
        if dtype.AEY in self.channels:
            idx = self._channels_indexes()[dtype.AEY]
            if isinstance(self.data, tuple):
                data = self.data[0] if len(self.data) == 1 else None
            else:
                data = self.data
            if data is not None:
                return data[:, idx]
        return None

    @property
    def TEY(self) -> npt.NDArray | None:
        """
        The Total Electron Yield (TEY) channel data if available.

        Returns
        -------
        npt.NDArray | None
            The TEY channel data if available. Otherwise None.
        """
        if dtype.TEY in self.channels:
            idx = self._channels_indexes()[dtype.TEY]
            if isinstance(self.data, tuple):
                data = self.data[0] if len(self.data) == 1 else None
            else:
                data = self.data
            if data is not None:
                return data[:, idx]
        return None



"""
Module for additional/adjacent functionality to add to pyNexafs.

Includes reduction, decorators, importing functions, and other utilities.
"""

# Importing modules:
from pyNexafs.utils import mda

# Decorators:
from pyNexafs.utils import decorators

# Reduction:
from pyNexafs.utils import reduction

# Filesizing
from pyNexafs.utils import sizes

__all__ = [
    "mda",
    "decorators",
    "reduction",
    "sizes",
]



"""
This module contains the `directory_selector` class.

This allows the selection of a filepath.
"""

import os
from PyQt6 import QtWidgets, QtCore, QtGui

# Extras for __main__ testing
from PyQt6.QtWidgets import QApplication
import sys


class directorySelector(QtWidgets.QWidget):
    """
    Module for selecting a directory path.

    Highlights the text

    Attributes
    ----------
    newPath : QtCore.pyqtSignal
        Generates a signal when a new path entry is created.
    folder_path:
    """

    new_path = QtCore.pyqtSignal(bool)
    """A signal to indicate the selection of a new (valid) path."""

    # String constants for widget elements
    edit_description = "Path to load NEXAFS data from."
    dialog_caption = "Select NEXAFS Directory"

    def __init__(self, parent: QtWidgets.QWidget | None = None):
        super().__init__(parent)

        layout = QtWidgets.QHBoxLayout()
        self.setLayout(layout)

        # Update margins to be compact if a parent is provided.
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            layout.setContentsMargins(0, 0, 0, 0)

        # Instance attributes
        self._folder_path = os.path.expanduser("~")  # default use home path.
        """The last valid directory selection"""

        self._folder_path = directorySelector.format_path(
            self._folder_path
        )  # adds final slash if not present.

        ### Instance widgets
        # Refresh button / toggle
        refresh_icon = QtGui.QIcon.fromTheme("view-refresh")
        self.refresh_button = QtWidgets.QPushButton(refresh_icon, "")
        self.refresh_button.clicked.connect(self.refresh_dir)
        # Folder path
        self.folder_path_edit = QtWidgets.QLineEdit()
        self.folder_path_edit.setText(self._folder_path)
        self.folder_path_edit.accessibleName = "Directory"
        self.folder_path_edit.accessibleDescription = self.edit_description
        self.folder_path_edit.editingFinished.connect(self.validate_edit_path)
        self.folder_path_edit_default_stylesheet = self.folder_path_edit.styleSheet()
        # Folder select button
        self.folder_select_button = QtWidgets.QPushButton("Browse")
        self.folder_select_button.clicked.connect(self.select_path)

        # Setup layout
        layout.addWidget(self.folder_path_edit)
        layout.addWidget(self.folder_select_button)
        layout.addWidget(self.refresh_button)

    def refresh_dir(self):
        """
        Re-emits a change signal for the current directory.
        """
        self.new_path.emit(True)

    @property
    def folder_path(self):
        """
        The direction selection of the selector widget.

        Parameters
        ----------
        path : str
            A directory string.

        Returns
        -------
        path : str
            The most recent valid directory.

        Raises
        ------
        ValueError
            If the
        """
        return self._folder_path

    @folder_path.setter
    def folder_path(self, path: str):
        # Hold old path
        old_path = self._folder_path
        # Validate the path
        v = self.validate_path(path)
        if v:
            formatted_path = self.format_path(path)
            # Check if the path is an update
            if formatted_path != old_path:
                # Update path
                self._folder_path = formatted_path
                # Update the UI
                self.folder_path_edit.setText(formatted_path)
                # Emit a path change
                self.new_path.emit(True)
            # Otherwise do nothing.
        else:
            raise ValueError(f"The path `{path}` does not exist.")

    def select_path(self) -> None:
        """
        Generates a file dialog to select a directory, then updates the path.
        """
        path = QtWidgets.QFileDialog.getExistingDirectory(
            parent=None,
            caption=self.dialog_caption,
            directory=self._folder_path,
            options=QtWidgets.QFileDialog.Option.ShowDirsOnly,
        )
        path = None if path == "" else path
        if path is None:
            # invalid path
            self.folder_path_edit.setText(
                self._folder_path
            )  # reset the text to the original path.
        else:
            # set editable path and validate.
            self.folder_path_edit.setText(path)
            self.validate_edit_path()

    @staticmethod
    def validate_path(path: str) -> bool:
        """
        Checks the validity of the selected path.

        Requires the path to exist.

        Parameters
        ----------
        path : str
            The path seeking to be used
        """
        return os.path.isdir(path)

    def validate_edit_path(self) -> None:
        """
        Validates the widget text edit path.

        Only updates internal path if the path is valid, otherwise sets background color to red.
        """
        # Get path text
        editable_path = self.format_path(self.folder_path_edit.text())

        # Check validity
        if editable_path is None:
            # invalid path
            self.folder_path_edit.setText(
                self.folder_path
            )  # reset the text to the original path.
            return

        if not self.validate_path(editable_path):
            # invalid path
            self.folder_path_edit.setStyleSheet("background-color: red;")
        else:
            # Valid path!
            # Reset the stylesheets.
            self.folder_path_edit.setStyleSheet(
                self.folder_path_edit_default_stylesheet
            )
            # Use the property to update the path and UI
            self.folder_path = editable_path

    @staticmethod
    def format_path(path: str) -> str:
        """
        Formats a path string to be consistent.

        Parameters
        ----------
        path : str
            The incoming path string. Must be a valid path, tested by os.path.isdir().

        Returns
        -------
        str
            A formatted path string that always contains a trailing slash, with slashes matching OS type.
        """

        # Strip whitespace, ensure tailing slash, and convert mixed slashes to forward slashes.
        formatted_path = os.path.join("", path.strip(), "").replace("\\", "/")
        # Remove redundant slash duplicates
        while "//" in formatted_path:
            formatted_path = formatted_path.replace("//", "/")
        # Convert slashes to match OS
        slashes = os.sep
        formatted_path = formatted_path.replace("/", slashes)
        # Add final slash if not present.
        if not formatted_path.endswith(slashes):
            formatted_path += slashes
        return formatted_path


if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = directorySelector()
    window.show()
    window.setWindowTitle("Directory Selector")
    sys.exit(app.exec())



"""
Table model and widget for the file display.
"""

import re
import traceback
import os
import sys
from typing import Any, override
import warnings
import numpy as np
from datetime import datetime as dt
from enum import Enum
from PyQt6 import QtWidgets, QtCore, QtGui
from PyQt6.QtWidgets import (
    QWidget,
    QTableView,
    QAbstractItemView,
    QProgressBar,
    QApplication,
    QStyle,
    QTableWidget,
    QTableWidgetItem,
)
from PyQt6.QtCore import (
    Qt,
    pyqtSignal,
    QSortFilterProxyModel,
    QRegularExpression,
    QAbstractTableModel,
    QThreadPool,
    # QThread,
)
from pyNexafs.parsers import parserBase
from pyNexafs.utils.sizes import btyes_to_human_readable
from pyNexafs.parsers.au.aus_sync.MEX2 import MEX2_NEXAFS


# Construct the Table Model
class tableModel(QAbstractTableModel):
    """
    Implements the QAbstractTableModel for a fileviewer

    Implements headersData to display provided headers,
    and the Adds two icons for load status (error, success)
    """

    class loadStatus(Enum):
        """
        The enumerate class for the load status of the file.
        """

        ERROR = 0
        HEADER_ONLY = 1
        FULL_DATA = 2

    _status_index = 0  # index for loaded status column.
    _status_header = ""  # string header for the loaded status column.

    def __init__(self, data, header=None):
        super(tableModel, self).__init__()
        self._data = data
        """A list of lists containing the data to be displayed in the table."""
        self._header = header

        # Initalise graphics for loaded / unloaded files.
        self._icon_error = QApplication.style().standardIcon(
            QStyle.StandardPixmap.SP_DialogCancelButton
        )
        """Icon for file load error status."""

        self._icon_header_only = QApplication.style().standardIcon(
            QStyle.StandardPixmap.SP_FileIcon
        )
        """Icon for file load header status."""

        self._icon_full_data = QApplication.style().standardIcon(
            QStyle.StandardPixmap.SP_DialogApplyButton
        )
        """Icon for file load success status."""

    @override
    def headerData(self, section, orientation, role):
        if self._header is None:
            return None
        if (
            orientation == Qt.Orientation.Horizontal
            and role == Qt.ItemDataRole.DisplayRole
        ):
            return self._header[section] if len(self._header) > section else None
        else:
            return super().headerData(section, orientation, role)

    @override
    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        # Check that second column is being indexed for "loaded" status, to display icons.
        if (
            self._header is not None
            and len(self._header) > self._status_index
            and self._header[self._status_index] == self._status_header
            and index.column() == self._status_index  # Is the status column.
        ):
            if role == Qt.ItemDataRole.DecorationRole:
                # Icons
                dat = self._data[index.row()][index.column()]
                if isinstance(dat, int):
                    return dat

                dat = self.loadStatus(dat)  # Convert to enum.
                match dat:
                    case self.loadStatus.ERROR:
                        val = self._icon_error
                    case self.loadStatus.HEADER_ONLY:
                        val = self._icon_header_only
                    case self.loadStatus.FULL_DATA:
                        val = self._icon_full_data
                    case _:
                        raise ValueError(f"Incorrect data match for {dat}")
                # val = (
                #     self._icon_full_data
                #     if dat
                #     else self._icon_error
                # )
                return val
            else:
                # val = "T" if self._data[index.row()][index.column()] else "F")
                return None

        # General data accessing
        if role == Qt.ItemDataRole.DisplayRole:
            return self._data[index.row()][index.column()]
        return None

    @override
    def rowCount(self, parent):
        # The length of the outer list.
        return len(self._data)

    @override
    def columnCount(self, parent):
        # The following takes the first sub-list, and returns
        # the length (only works if all rows are an equal length)
        if (
            hasattr(self._data, "__len__")
            and len(self._data) > 0
            and hasattr(self._data[0], "__len__")
        ):
            return len(self._data[0])
        else:
            return 0


class directoryViewerTable(QTableView):
    """
    A table widget that displays pyNexafs related parser files.

    Attributes:
    selection_loaded : pyqtSignal
        Emits after a new selection is made.
    __default_headers_list : list
        A
    """

    selection_loaded = pyqtSignal(bool)
    """A signal for emitting after selection change is loaded."""

    __default_headers_list: list[str] = [
        "#",
        "Filename",
    ]  # index and filename #TODO: Add "Created", "Modified" when implemented in parserBase.

    def __init__(
        self,
        init_dir: str | None = None,
        progress_bar: QProgressBar | None = None,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)

        # Instance properties
        self.files: list[str] = []
        """The list of parser filenames"""
        self._header_names: list[str] | None = None
        """Header names populated by the parser class `parser.SUMMARY_PARAM_NAMES`"""
        self._header_names_custom: list[str] | None = None
        """A custom set of parameters that can be set by setting the `headers` property.
        Overrides use of `directoryViewerTable._header_names`."""

        ## WE KEEP A DICT OF PARSERS, AND THEN USE THE PARSERS TO GENERATE THE TABLE.
        self._parser_headers: dict[str, parserBase] = {}
        self._parser_files: dict[str, parserBase] = {}

        self._viewing_directory: str | None = None
        self._parser: parserBase | None = None
        self._filetype_filters: list[str] | None = None
        self._str_filter: str | None = None

        # Instance Widgets
        self.files_model = None
        self.proxy_model = QSortFilterProxyModel()

        # Instance Attributes
        self.verticalHeader().hide()  # remove default row numbers
        self.setModel(self.proxy_model)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSortingEnabled(True)  # enabled by proxymodel.
        self.proxy_model.setSourceModel(None)
        self.proxy_model.setFilterRegularExpression(None)
        self.proxy_model.setFilterKeyColumn(
            -1
        )  # use all columns for filtering instead of just first.
        self.proxy_model.sort(0, Qt.SortOrder.AscendingOrder)
        # self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
        self.resizeColumnToContents(0)
        self.setMinimumWidth(1)
        self.horizontalHeader().setMinimumSectionSize(10)

        # Event connections
        self.selectionModel().selectionChanged.connect(self.load_selection)

        # Progress bar
        self._progress_bar = progress_bar

        # Initialise viewing directory
        if init_dir is not None:
            self.directory = init_dir

        # Setup default header columns if parser is defined. Use header/index definition in table_model.
        self.__default_parsed_headers_list = self.__default_headers_list.copy()
        self.__default_parsed_headers_list.insert(
            self._status_index(),
            self._status_header(),  # At 1st index, insert ""
        )

    @classmethod
    def _status_index(cls) -> int:
        """
        Returns the index of the loaded status column, an attribute of the abstract table model class.
        """
        return tableModel._status_index

    @classmethod
    def _status_header(cls) -> str:
        """
        Returns the string header of the loaded status column, an attribute of the abstract table model class.
        """
        return tableModel._status_header

    @property
    def _default_headers(self) -> list[str]:
        """
        Returns a copy of the default headers depending on if a parser is defined (adds a loading column).

        Returns
        -------
        list[str]
            List of the default headers
        """
        obj = (
            self.__default_headers_list
            if self.parser is None
            else self.__default_parsed_headers_list
        )
        return obj.copy()

    @property
    def header(self) -> list[str] | None:
        """
        Returns the current header list for selected parsed.

        Parameters
        ----------
        val : list[str] | None
            Sets a custom list of headers to use for the table summary.

        Returns
        -------
        list[str]
            A list of column names currently observed by the parser.
        """
        return (
            self._header_names_custom
            if self._header_names_custom is not None
            else self._header_names
        )

    @header.setter
    def header(self, val: list[str] | None):
        if val != self._header_names_custom:
            if val is None:
                self._header_names_custom = None
            else:
                self._header_names_custom = self._default_headers + val

            # Update the table with the new headers.
            self.update_header()

    @header.deleter
    def header(self):
        self._header_names_custom = None

    @property
    def parser(self) -> type[parserBase] | None:
        """
        Returns the currently selected parser.

        Returns
        -------
        type[parserBase] | None
            The currently selected parser, or None if no parser is selected.
        """
        return self._parser

    @parser.setter
    def parser(self, parser: type[parserBase] | None):
        """
        Set the parser, used for tabling header information.

        Resets the directory_viewer_table._file_params and directory_viewer_table.filedata dictionaries.

        Parameters
        ----------
        parser : type[parserBase] | None
            The currently selected parser, or None if no parser is selected.
        """
        self._parser = parser
        # Reset existing file data to use new parser.
        self._parser_files = {}
        self._parser_headers = {}
        self._filetype_filters = (
            parser.ALLOWED_EXTENSIONS if parser is not None else None
        )

        self.update_table()

    @parser.deleter
    def parser(self):
        """
        Remove the parser from the directory viewer.
        """
        self._parser = None
        self._parser_files = {}
        self._parser_headers = {}
        self.update_table()

    @property
    def directory(self):
        """
        Returns the currently viewed directory.

        Returns
        -------
        str
            The currently viewed directory.
        """
        return self._viewing_directory

    @directory.setter
    def directory(self, target: str):
        """
        Set the directory to view.

        Only updates the files list if directory is changed.
        """

        # Check if same directory.
        if self._viewing_directory == target:
            return

        # Get new file list
        self.files = [
            file
            for file in os.listdir(target)
            if os.path.isfile(os.path.join(target, file))
        ]
        # Clear existing data
        self._parser_headers = {}
        self._parser_files = {}

        # Update directory
        self._viewing_directory = target

        # Push update to table.
        self.update_table()

    @property
    def filters(self) -> tuple[str | None, list[str] | None]:
        """
        Returns the currently set filters.

        Returns
        -------
        tuple[str | None, list[str] | None]
            A tuple containing the filename filter and the filetype filter.
        """
        return (self._str_filter, self._filetype_filters)

    @filters.setter
    def filters(self, filters: tuple[str | None, list[str] | None]):
        """
        Set the filters for the directory viewer.

        Parameters
        ----------
        filters : tuple[str | None, list[str] | None]
            A tuple containing the filename filter and the filetype list filter.
        """
        str_filter, filetype_filters = filters

        # Correct type for empty strings.
        if str_filter == "":
            str_filter = None

        # Check if str filter is changed:
        if str_filter != self._str_filter:
            self._str_filter = str_filter
            # Update proxy model
            QRegExp = QRegularExpression(
                str_filter, QRegularExpression.PatternOption.CaseInsensitiveOption
            )
            self.proxy_model.setFilterRegularExpression(
                None if str_filter is None else QRegExp
            )
            # self.proxy_model.setFilterFixedString(str_filter)

        # Check if filetype filter is changed:
        if filetype_filters is not None and isinstance(filetype_filters, list):
            # Check if filetype filter is changed:
            if (
                self._filetype_filters is not None
                and len(self._filetype_filters) == len(filetype_filters)
                and np.all(
                    [
                        (
                            filetype_filters[i] == self._filetype_filters[i]
                            or filetype_filters[i] in self._filetype_filters
                        )
                        for i in range(len(filetype_filters))
                    ]
                )
            ):
                # Everything the same...
                pass
            else:
                self._filetype_filters = filetype_filters
                self.update_table()
        elif filetype_filters is None:
            self._filetype_filters = (
                self._parser.ALLOWED_EXTENSIONS if self._parser is not None else None
            )
        else:
            raise AttributeError(
                f"filters[1] '{filetype_filters}' is not a list of filetype strings or None."
            )

    @filters.deleter
    def filters(self):
        """
        Remove the filters for the directory viewer.
        """
        self._str_filter, self._filetype_filters = None, None
        self.update_table()

    @property
    def progress_bar(self) -> QProgressBar:
        """
        Returns an associated progress bar widget if provided, else None.

        Parameters
        ----------
        progress_bar: QProgressBar
            The progress bar widget.

        Returns
        -------
        QProgressBar
            The progress bar widget.
        """
        return self._progress_bar

    @progress_bar.setter
    def progress_bar(self, progress_bar: QProgressBar):
        self._progress_bar = progress_bar

    def update_header(self) -> None:
        """
        Generates the column headers using default parser important parameters.

        First attempts to call `summary_param_names_with_units` on the first non-None
        header-loaded parser object, to load the column labels and their appropriate units.
        Alternatively calls `summary_param_names` on the object, which doesn't require unit labels.
        These two methods take into account the parser `relabel` property.
        If no headers are loaded then calls the parser `summary_param_names`.
        A change in header will also trigger a change in the table model.
        """
        # Setup default header columns
        header = self._default_headers  # Get defaults, i.e. (loaded), number and name.

        # Add list values if parser is defined, and header objects allow unit calls.
        if self.parser is not None and len(self._parser_headers) > 0:
            val = None
            # Collect the first non-None parser object.
            for val in self._parser_headers.values():
                # Collect val (parser objects), where val is not None.
                if val is not None:
                    assert isinstance(val, self.parser)
                    break  # end for loop prematurely.
            # Establish parameter header names:
            if val is not None:
                try:
                    # Attempt to use 'summary_param_names_with_units' if available.
                    obj_head = val.summary_param_names_with_units
                except NotImplementedError:
                    warnings.warn(
                        f"{self.parser.__name__} has not implemented 'summary_param_names_with_units'. Defaulting to 'summary_param_names'."
                    )
                    obj_head = val.summary_param_names  # without units.
            else:
                # If all parser headers are None, then use parser default values.
                obj_head = self.parser.summary_param_names
            # Add obj_head to header
            header += [s[0] if isinstance(s, tuple) else s for s in obj_head]
        # If _parser_headers are not defined but the parser is.
        elif self.parser is not None:
            header += [
                s[0] if isinstance(s, tuple) else s
                for s in self.parser.summary_param_names
            ]

        # Update internal header names
        self._header_names = header

        # Ideally just updates header names, but occasionally crashes when attempting to layoutChange.emit()
        if self.files_model is not None:
            old_selection = self.selectionModel().selectedRows()
            new_model = tableModel(
                data=self.files_model._data,
                header=(
                    self._header_names
                    if self._header_names_custom is None
                    else self._header_names_custom
                ),
            )
            self.proxy_model.setSourceModel(new_model)
            self.files_model = new_model
            # Setup and restore selection: requires multiselection to select each row before using regular extended selection.
            self.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
            [self.selectRow(i.row()) for i in old_selection]
            self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        # Update column widths with header update.
        for i in range(0, len(self._header_names)):
            # If the current column width is larger than the size hint, resize to fit for all columns.
            if self.columnWidth(i) > self.sizeHintForColumn(i):
                self.resizeColumnToContents(i)
        return

    def update_table(self):
        """
        Update the directory view.

        Applies text filter, parser and permissionable extensions to filter the files list.
        Loads the filtered header data for the table, appling updates to the class progress bar if defined.

        """
        # Do nothing if an empty directory.
        if self.directory is None:
            return

        ### COLLECT FILES AND APPLY FILETYPE FILTER
        # Collect available files
        files = self.files
        # 1: Filetypes
        if self._filetype_filters is not None and len(self._filetype_filters) > 0:
            files = [
                file
                for file in self.files
                if file.endswith(tuple(self._filetype_filters))
            ]
        # Set the progress bar to zero.
        if self.progress_bar is not None:
            self.progress_bar.setValue(0)
            # Store number of files to be processed for progress bar
            self.progress_bar.setRange(0, len(files) if len(files) > 0 else 1)
        # 2: Load new fileheader data.
        if self.parser is not None:
            # Get header data
            with warnings.catch_warnings(record=True) as recorded_warnings:
                for i, file in enumerate(files):
                    if file not in self._parser_headers:
                        try:
                            self._parser_headers[file] = self.parser(
                                os.path.join(self.directory, file), header_only=True
                            )
                        # Catch unimplemented and import errors.
                        except (NotImplementedError, ImportError, PermissionError):
                            self._parser_headers[file] = None
                    if self.progress_bar is not None:
                        # self.progress_bar.setValue(int((i + 1) / files_len * 100))
                        self.progress_bar.setValue(i + 1)
            # Process the caught warnings
            warning_categories = {}
            warn_re = re.compile(
                r"Attempted method '(.*)' failed to load '(.*)' from '(.*)' with (.*)."
            )
            for w in recorded_warnings:
                matches = warn_re.search(str(w.message))
                if matches and len(matches.groups()) == 4:
                    method, file, parser, error = matches.groups()
                    # Collect warning
                    parser_method = f"{parser}.{method}"
                    filetype = file.split(".")[-1]
                    filetype_error = f"{filetype}:{error}"

                    if parser_method not in warning_categories:
                        # Add parser_method to categories
                        warning_categories[parser_method] = {filetype_error: 1}
                    elif filetype_error not in warning_categories[parser_method]:
                        # Add filetype_error to parser_method
                        warning_categories[parser_method][filetype_error] = 1
                    else:
                        # Increment filetype_error count
                        warning_categories[parser_method][filetype_error] += 1
                else:
                    # re-issue warning
                    warnings.warn_explicit(
                        message=w.message,
                        category=w.category,
                        filename=w.filename,
                        lineno=w.lineno,
                        source=w.source,
                    )
            # Print warning categories
            for parser_method, file_errors in warning_categories.items():
                for filetype_error, count in file_errors.items():
                    filetype, error = filetype_error.split(":")
                    print(
                        f"Parser method '{parser_method}' failed {count} times on filetype '{filetype}' with error '{error}'."
                    )
        else:
            # If parser is None, empty the existing data.
            self._parser_headers = {}

        # Update headers
        self.update_header()

        # Get header data for table:
        data = []  # list of file header lists.
        excess = [
            "" for _ in range(len(self._header_names) - len(self._default_headers))
        ]  # empty columns for invalid file headers.
        for i, file in enumerate(files):
            filedata = [i + 1, file]  # add index (1, ...) and filename to data.
            # If parser specified (for loading) add field for successful loading.
            if self.parser:
                status = (
                    True
                    if file in self._parser_headers
                    and self._parser_headers[file] is not None
                    else False
                )
                filedata.insert(self._status_index(), status)

            # Raise an error if mismatch between info and header variables.
            if len(filedata) != len(self._default_headers):
                raise AttributeError("Default headers do not match filedata.")

            # Add parameter values
            if file in self._parser_headers and isinstance(
                self._parser_headers[file], parserBase
            ):
                # Match indexing of parser.SUMMARY_PARAM_RAW_NAMES
                filedata.extend(self._parser_headers[file].summary_param_values)
            else:
                filedata.extend(excess)
            # Add header data to data list.
            data.append(filedata)

        # New table - Ideally would be a model update, but would crash when attempting to layoutChange.emit()
        self.files_model = tableModel(data=data, header=self._header_names)
        # Update proxy (filtering / sorting) model
        self.proxy_model.setSourceModel(self.files_model)

        self.resizeColumnToContents(0)
        for i in range(0, len(self._header_names)):
            # If the current column width is larger than the size hint, resize to fit for all columns.
            if self.columnWidth(i) > self.sizeHintForColumn(i):
                self.resizeColumnToContents(i)
        if self.progress_bar is not None:
            self.progress_bar.setValue(0)
        return

    def load_selection(self) -> None:
        """
        Loads full data from files of the current selection.

        Instead of just header information, loads all data lines
        from the selected files. Loaded datafiles are cached in
        self._parser_files, until the parser object is changed.
        """
        sm = self.selectionModel()
        # Require selection and parser to trigger load data.
        if sm.hasSelection() and self.parser is not None:
            rows = sm.selectedRows(
                column=1 if self._status_index() > 1 else 2
            )  # get filename, not index.

            # Set the progress bar to zero.
            if self.progress_bar is not None:
                self.progress_bar.setValue(0)
                # Store number of files to be processed for progress bar
                self.progress_bar.setRange(0, len(rows))

            # Setup warnings to show once.
            warnings.simplefilter("once", UserWarning)
            for i, row in enumerate(rows):
                filename = self.proxy_model.data(row, Qt.ItemDataRole.DisplayRole)
                # Check if the file already loaded:
                if (
                    filename not in self._parser_files
                    or self._parser_files[filename] is None
                ):
                    # Check for a header object (should always exist).
                    if (
                        filename in self._parser_headers
                        and self._parser_headers[filename] is not None
                    ):
                        # Use existing parser instance, add to files.
                        parser = self._parser_headers[filename]
                        assert isinstance(parser, parserBase)
                        try:
                            parser.load()  # use internal filepath to load
                            if parser.is_loaded:
                                self._parser_files[filename] = parser
                        except Exception as e:
                            traceback.print_exception(e)
                    # else: ignore unloaded headers (None) in the selection.
                if self.progress_bar is not None:
                    self.progress_bar.setValue(i + 1)
            # Restore warnings functionality.
            warnings.resetwarnings()
            # Trigger a loading completed signal.
            self.progress_bar.setValue(0)
            self.selection_loaded.emit(True)
        return

    @property
    def selected_filenames(self) -> list[str]:
        """
        Returns the selected filenames in the directory viewer.

        Returns
        -------
        list[str]
            A list of filenames selected in the directory viewer.
        """
        sm = self.selectionModel()
        if sm.hasSelection():
            rows = sm.selectedRows(
                column=self._index_filename_column
            )  # Filename in second column.
            # return self.files_model.data(rows, Qt.ItemDataRole.DisplayRole)
            return [
                self.proxy_model.data(row, Qt.ItemDataRole.DisplayRole) for row in rows
            ]
        return []

    @property
    def _index_filename_column(self) -> int:
        """
        Returns the index of the filename column in the table model.
        """
        if self.parser is None:
            return 1
        else:
            return 2 if self._status_index() < 2 else 1


# TODO: Implement threading for file header/data loading
# class fileLoaderWorker(QtCore.QRunnable):
#     """
#     Class to load files in a separate thread.

#     Parameters
#     ----------
#     n : int
#         The thread number.
#     files : list[str]
#         The list of files to load.
#     parser : type[parserBase]
#         The parser class to use for loading the files.
#     parsers: dict[str, parserBase] | None
#         The dict of existing parsers to use for loading the files.
#         I.e. If a header has been loaded but not the full data, use
#         the same parser to load the data.
#     QProgressBar : QProgressBar
#         The progress bar to update during loading.
#     """
#     def __init__(self, n: int,
#                  files: list[str],
#                  parser: type[parserBase],
#                  parsers: dict[str, parserBase] | None = None,
#                  progress_bar: QProgressBar | None = None,
#                  load_kwargs: dict[str, Any] | None = None):
#         super().__init__()
#         self.n = n
#         self.files = files
#         self.parser = parser
#         self.load_kwargs = load_kwargs
#         self.progress_bar = progress_bar

#         # Store the parsers if provided.
#         self.parsers = parsers
#         """
#         The dict of provided parsers in the thread.
#         """

#         # Create storage for loaded parsers.
#         self.loaded: dict[str, parserBase] = {}
#         """
#         The dict of parsers loaded in the thread.
#         """

#     def run(self):
#         """
#         Load the files in a separate thread.

#         This method is called when the thread is started.
#         """
#         # Load the files using the parser class.
#         for i, file in enumerate(self.files):
#             if file not in self.loaded or not isinstance(self.loaded[file], self.parser):
#                 try:
#                     # Check if the file already has a parser (i.e. header).
#                     if self.parsers is not None and file in self.parsers:
#                         # Use the provided parser if it exists.
#                         parser = self.parsers[file]
#                         parser.load(**(self.load_kwargs
#                                        if self.load_kwargs is not None
#                                        else {}))
#                     else:
#                         parser = self.parser(file,
#                                             **(self.load_kwargs
#                                                if self.load_kwargs is not None
#                                                else {}))
#                     if self.progress_bar is not None:
#                         self.progress_bar.setValue(i + 1)
#                     # Add the loaded parser to the dictionary.
#                     self.loaded[file] = parser
#                 except Exception as e:
#                     traceback.print_exception(e)
#                     if self.progress_bar is not None:
#                         self.progress_bar.setValue(self.progress_bar.value() + 1)
#             else:
#                 # If the file is already loaded, just update the progress bar.
#                 if self.progress_bar is not None:
#                     self.progress_bar.setValue(self.progress_bar.value() + 1)

# class fileLoaderCounter(QObject):
#     """
#     Counts the completion of
#     """


class directoryViewerTableNew(QTableView):
    """
    A table widget that displays pyNexafs related parser files.

    This TableView operates on the following properties.
    - `parser`: This is the parser class that is used to (try to) load the files.
    - `directory`: This is the directory that is being viewed.
    - `filters`: This is a tuple of the filename filter and the filetype filter.
    - `progress_bar`: This is a progress bar that is used to show the progress of loading files.
    - `header` : The parser default param names (default), but can be a selected list of parameter names.
            Uses the `_header_param_names_default` and `_header_param_names_custom` attributes to determine
            the header names (which are set by ).

    Attributes:
    selection_loaded : pyqtSignal
        Emits after a new selection is made.
    __default_headers_list : list
        A

    See Also
    --------
    SummaryParamSelector
        A widget that allows the user to select the summary parameters to be displayed in the table.
    """

    selection_loaded = pyqtSignal(bool)
    """A signal for emitting after selection change is loaded."""

    __REQUIRED_HEADER_LIST: list[str] = ["#", "Filename"]
    """Headers always created for the table: file number and filename."""

    __DEFAULT_FILE_HEADERS: list[str] = ["created", "modified", "memory_size"]
    """When no parser selected, default file info headers."""

    __DEFAULT_PARSER_HEADER_KEYS: list[str] = ["created", "modified", "memory_size"]
    """Default parser parameter values selected upon parser selection, in addition to SUMMARY_PARAM_NAMES."""

    def __init__(
        self,
        init_dir: str | None = None,
        progress_bar: QProgressBar | None = None,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)

        # Instance properties
        self._files: list[str] = []
        """The list of directory filenames"""

        self._header_param_names_default: list[str] | None = None
        """Header names populated by `__DEFAULT_HEADER_KEYS` the parser class `parser.SUMMARY_PARAM_NAMES`"""
        self._header_param_names_custom: list[str] | None = None
        """A custom set of parameters that can be set by setting the `headers` property.
        Overrides use of `directoryViewerTable._header_names`."""

        ## WE KEEP A DICT OF PARSERS, AND THEN USE THE PARSERS TO GENERATE THE TABLE.
        self._parser_headers: dict[str, parserBase | None] = {}
        """The parsers who have been loaded with headers only."""
        self._parser_files: dict[str, parserBase | None] = {}
        """The parsers who have been loaded with full data."""

        self._viewing_directory: str | None = None
        """The internal directory that is being viewed."""
        self._parser: type[parserBase] | None = None
        """The selected parser class."""
        self._filetype_filters: list[str] | None = None
        """The list of filetypes that are allowed to be loaded."""
        self._str_filter: str | None = None
        """The string filter that is used to filter the filenames. Can also be QRegex"""

        # Instance Widgets
        self.files_model = None
        """The model for the table, before filtering."""
        self.proxy_model = QSortFilterProxyModel()
        """The filtered model for the table, using `self.files_model` as the source model."""

        # Threading

        # Instance Attributes
        self.verticalHeader().hide()  # remove default row numbers
        self.setModel(self.proxy_model)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSortingEnabled(True)  # enabled by proxymodel.
        self.proxy_model.setSourceModel(None)
        self.proxy_model.setFilterRegularExpression(None)
        self.proxy_model.setFilterKeyColumn(
            -1
        )  # use all columns for filtering instead of just first.
        self.proxy_model.sort(0, Qt.SortOrder.AscendingOrder)
        # self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
        self.resizeColumnToContents(0)
        self.setMinimumWidth(1)
        self.horizontalHeader().setMinimumSectionSize(10)

        # Event connections
        self.selectionModel().selectionChanged.connect(self.load_selection)

        # Progress bar
        self._progress_bar = progress_bar

        # Initialise viewing directory
        if init_dir is not None:
            self.directory = init_dir

        self.update_header()
        self.update_table()

    @property
    def files(self) -> list[str]:
        """
        Returns the list of parser files in the directory.

        Subsets the filelist to only include files that
        are within the `parser.ALLOWED_EXTENSIONS` and `self._filetype_filters`.

        Returns
        -------
        list[str]
            A list of filenames in the directory.
        """
        # Is a parser defined? If not, return all files.
        if self.parser is None:
            return self._files
        # Check if a parser filetype filter has been selected
        elif self._filetype_filters is not None and len(self._filetype_filters) > 0:
            return [
                file
                for file in self._files
                if file.endswith(tuple(self._filetype_filters))
            ]
        # Check if a parser is selected but no filetype filter is defined.
        # though usually this is not the case.
        elif len(self.parser.ALLOWED_EXTENSIONS) > 0:
            return [
                file
                for file in self._files
                if file.endswith(tuple(self.parser.ALLOWED_EXTENSIONS))
            ]
        else:
            return self._files

    @classmethod
    def _status_index(cls) -> int:
        """
        Returns the index of the loaded status column, an attribute of the abstract table model class.
        """
        return tableModel._status_index

    @classmethod
    def _status_header(cls) -> str:
        """
        Returns the string header of the loaded status column, an attribute of the abstract table model class.
        """
        return tableModel._status_header

    @property
    def header(self) -> list[str] | None:
        """
        Returns the current header list for selected parsed.

        Includes values from `__REQUIRED_HEADER_LIST`, if a parser is defined or not.

        Parameters
        ----------
        val : list[str] | None
            Sets a custom list of headers to use for the table summary.

        Returns
        -------
        list[str]
            A list of column names currently observed by the parser.
        """
        return (
            self._header_param_names_custom
            if self._header_param_names_custom is not None
            else self._header_param_names_default
        )

    @header.setter
    def header(self, val: list[str] | None):
        if val != self._header_param_names_custom:
            if val is None:
                self._header_param_names_custom = None
            else:
                self._header_param_names_custom = self.__REQUIRED_HEADER_LIST + [
                    v for v in val if v not in self.__REQUIRED_HEADER_LIST
                ]

            # Update the table with the new headers.
            self.update_header()

    @header.deleter
    def header(self):
        self._header_param_names_custom = None

    # @property
    # def header_optional(self) -> list[str] | None:
    #     """
    #     The optional selection for the current header list for selected parser.

    #     Excludes values from `__REQUIRED_HEADER_LIST`.

    #     Returns
    #     -------
    #     list[str]
    #         A list of column names currently observed by the parser.
    #     """
    #     h = (self._header_param_names_custom
    #             if self._header_param_names_custom is not None
    #             else self._header_param_names_default
    #     )
    #     return

    @property
    def parser(self) -> type[parserBase] | None:
        """
        Property for the selected parser.

        Parameters
        ----------
        parser : type[parserBase] | None
            A new parser class to use for loading files.
            The parser is used to set default extensions, and also
            resets the file data and header data and updates the table.

        Returns
        -------
        type[parserBase] | None
            The currently selected parser, or None if no parser is selected.
        """
        return self._parser

    @parser.setter
    def parser(self, parser: type[parserBase] | None):
        if self._parser != parser:
            self._parser = parser
            # Reset existing file data to use new parser.
            self._parser_files = {}
            self._parser_headers = {}
            self._filetype_filters = (
                parser.ALLOWED_EXTENSIONS if parser is not None else None
            )

            # Load the parser headers for the current directory.
            self._collect_parser_headers()

            # Generate a new header from the parser class.
            self._header_param_names_default = self.generate_header()
            print(self._parser, "New header: ", self._header_param_names_default)

            # Reset the custom header names.
            self._header_param_names_custom = None

            # Generate a new header from the loaded parsers and update headers.
            # self.header = self.generate_header()
            self.update_header()

            # Update the table with the new parser.
            self.update_table()

    @parser.deleter
    def parser(self):
        self._parser = None
        self._parser_files = {}
        self._parser_headers = {}
        self._header_param_names_default = None
        self._header_param_names_custom = None
        self.update_header()
        self.update_table()

    @property
    def directory(self):
        """
        Returns the currently viewed directory.

        Returns
        -------
        str
            The currently viewed directory.
        """
        return self._viewing_directory

    @directory.setter
    def directory(self, target: str):
        """
        Set the directory to view.

        Only updates the files list if directory is changed.
        """

        # Check if same directory.
        if self._viewing_directory == target:
            return

        # Get new file list
        self._files = [
            file
            for file in os.listdir(target)
            if os.path.isfile(os.path.join(target, file))
        ]
        # Clear existing data
        self._parser_headers = {}
        self._parser_files = {}

        # Update directory
        self._viewing_directory = target

        # Load the new parsers if a parser is defined.
        if self.parser is not None:
            self._collect_parser_headers()

        # Push update to table.
        self.update_table()

    def _collect_parser_headers(self) -> None:
        """
        Load the headers for the current directory if the
        parsers are not loaded already (allows refreshing).

        Updates `self._parser_headers` with the loaded parsers.
        Also catches warnings.

        Raises
        -------
        AttributeError
            If no parser is defined.
        """
        if self.parser is None:
            raise AttributeError("No parser defined. Cannot collect parser headers.")
        if self.directory is None:
            raise AttributeError("No directory defined. Cannot collect parser headers.")
        # Set the progress bar to zero.
        files = self.files
        if self.progress_bar is not None:
            self.progress_bar.setValue(0)
            # Store number of files to be processed for progress bar
            self.progress_bar.setRange(0, len(files) if len(files) > 0 else 1)
        with warnings.catch_warnings(record=True) as recorded_warnings:
            for i, file in enumerate(files):
                # Check file not loaded already, or parser is not current type.
                if file not in self._parser_headers or (
                    self.parser is not None
                    and not isinstance(self._parser_headers[file], self.parser)
                ):
                    # Attempt to load
                    try:
                        self._parser_headers[file] = self.parser(
                            os.path.join(self.directory, file), header_only=True
                        )
                    except (NotImplementedError, ImportError, PermissionError):
                        self._parser_headers[file] = None
                if self.progress_bar is not None:
                    # self.progress_bar.setValue(int((i + 1) / files_len * 100))
                    self.progress_bar.setValue(i + 1)
            # When finished
            if self.progress_bar is not None:
                self.progress_bar.setValue(0)
        # Process the caught warnings
        warning_categories = {}
        warn_re = re.compile(
            r"Attempted method '(.*)' failed to load '(.*)' from '(.*)' with (.*)."
        )
        for w in recorded_warnings:
            matches = warn_re.search(str(w.message))
            if matches and len(matches.groups()) == 4:
                method, file, parser, error = matches.groups()
                # Collect warning
                parser_method = f"{parser}.{method}"
                filetype = file.split(".")[-1]
                filetype_error = f"{filetype}:{error}"

                if parser_method not in warning_categories:
                    # Add parser_method to categories
                    warning_categories[parser_method] = {filetype_error: 1}
                elif filetype_error not in warning_categories[parser_method]:
                    # Add filetype_error to parser_method
                    warning_categories[parser_method][filetype_error] = 1
                else:
                    # Increment filetype_error count
                    warning_categories[parser_method][filetype_error] += 1
            else:
                # re-issue warning
                warnings.warn_explicit(
                    message=w.message,
                    category=w.category,
                    filename=w.filename,
                    lineno=w.lineno,
                    source=w.source,
                )
        # Print warning categories
        for parser_method, file_errors in warning_categories.items():
            for filetype_error, count in file_errors.items():
                filetype, error = filetype_error.split(":")
                print(
                    f"Parser method '{parser_method}' failed {count} times on filetype '{filetype}' with error '{error}'."
                )

    @property
    def filters(self) -> tuple[str | None, list[str] | None]:
        """
        The current filters (filename and filetype) for the directory viewer.

        The filename filter is a string that is used to filter the filenames in the directory.
        Regular expressions
        The

        Parameters
        ----------
        filters : tuple[str | None, list[str] | None]
            A tuple containing the filename filter and the filetype list filter.

        Returns
        -------
        tuple[str | None, list[str] | None]
            A tuple containing the filename filter and the filetype filter.
        """
        return (self._str_filter, self._filetype_filters)

    @filters.setter
    def filters(self, filters: tuple[str | None, list[str] | None]):
        str_filter, filetype_filters = filters

        # Correct type for empty strings.
        if str_filter == "":
            str_filter = None

        # Check if str filter is changed:
        if str_filter != self._str_filter:
            self._str_filter = str_filter
            # Update proxy model
            QRegExp = QRegularExpression(
                str_filter, QRegularExpression.PatternOption.CaseInsensitiveOption
            )
            self.proxy_model.setFilterRegularExpression(
                None if str_filter is None else QRegExp
            )
            # self.proxy_model.setFilterFixedString(str_filter)

        # Check if filetype filter is changed:
        if filetype_filters is not None and isinstance(filetype_filters, list):
            # Check if filetype filter is changed:
            if (
                self._filetype_filters is not None
                and len(self._filetype_filters) == len(filetype_filters)
                and np.all(
                    [
                        (
                            filetype_filters[i] == self._filetype_filters[i]
                            or filetype_filters[i] in self._filetype_filters
                        )
                        for i in range(len(filetype_filters))
                    ]
                )
            ):
                # Everything the same...
                pass
            else:
                self._filetype_filters = filetype_filters
                self.update_table()
        elif filetype_filters is None:
            self._filetype_filters = (
                self.parser.ALLOWED_EXTENSIONS if self.parser is not None else None
            )
        else:
            raise AttributeError(
                f"filters[1] '{filetype_filters}' is not a list of filetype strings or None."
            )

    @filters.deleter
    def filters(self):
        """
        Remove the filters for the directory viewer.
        """
        self._str_filter, self._filetype_filters = None, None
        self.update_table()

    @property
    def filter_filename(self) -> str | None:
        """
        Returns the current filename filter.

        Returns
        -------
        str | None
            The current filename filter.
        """
        return self._str_filter

    @property
    def progress_bar(self) -> QProgressBar | None:
        """
        Returns an associated progress bar widget if provided, else None.

        Parameters
        ----------
        progress_bar: QProgressBar
            The progress bar widget.

        Returns
        -------
        QProgressBar | None
            The progress bar widget.
        """
        return self._progress_bar

    @progress_bar.setter
    def progress_bar(self, progress_bar: QProgressBar):
        self._progress_bar = progress_bar

    ### Functions to collect information to build and update the table model from the selected directory, parser and filters.
    def generate_header(self) -> list[str]:
        """
        Generate the column headers using the default parser important parameters.

        First attempts to call `summary_param_names_with_units` on the first non-None
        header-loaded parser object, to load the column labels and their appropriate units.
        Alternatively calls `summary_param_names` on the object, which doesn't require unit labels.
        These two methods take into account the parser `relabel` property.
        """
        # Setup default header columns
        header = (
            self.__REQUIRED_HEADER_LIST.copy()
        )  # Get the default headers, but don't modify the original list.
        # Add list values if parser is defined, and header objects allow unit calls.
        if not (self.parser is not None and len(self._parser_headers) > 0):
            # No parser or no headers (of files), so use default file headers.
            header += (
                self.__DEFAULT_FILE_HEADERS
            )  # Get defaults, i.e. (created, modified, size).
        else:
            header += (
                self.__DEFAULT_PARSER_HEADER_KEYS
            )  # Get defaults, i.e. (loaded), number and name.
            # Add the status header i.e. "Load Success"
            header.insert(self._status_index(), self._status_header())

            # Collect the first non-None parser object.
            val = None
            # Collect the first non-None parser object.
            for val in self._parser_headers.values():
                # Collect val (parser objects), where val is not None.
                if val is not None:
                    assert isinstance(val, self.parser)
                    break
            # Establish parameter header names:
            if val is not None:
                try:
                    # Attempt to use 'summary_param_names_with_units' if available.
                    obj_head = val.summary_param_names_with_units
                except NotImplementedError:
                    warnings.warn(
                        f"{self.parser.__name__} has not implemented 'summary_param_names_with_units'. Defaulting to 'summary_param_names'."
                    )
                    obj_head = val.summary_param_names  # without units.
            else:
                # If all parser headers are None, then use parser default values.
                obj_head = self.parser.summary_param_names

            # Add obj_head to header
            header += obj_head

        return header

    def generate_header_data(self) -> list[list[Any]]:
        """
        Collect the data corresponding to the header names.

        Uses the parser objects to generate the header data.
        """
        # Do nothing if an empty directory.
        if self.directory is None:
            return [[]]

        TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

        path = self.directory
        files = self.files
        parsers = self._parser_headers

        data = [[] for _ in range(len(files))]  # list of file header lists.
        if self.parser is None:
            # Get defaults, matching `self.__REQUIRED_HEADER_LIST` and `self.__DEFAULT_FILE_HEADERS`
            # - "#", "Filename", "Created", "Modified", "Size"
            for i, file in enumerate(files):
                data[i] = [i + 1, file]  # add index (1, ...) and filename to data.
                # Use OS to get file info
                data[i].extend(
                    [
                        dt.fromtimestamp(
                            os.path.getctime(os.path.join(path, file))
                        ).strftime(TIME_FORMAT),
                        dt.fromtimestamp(
                            os.path.getmtime(os.path.join(path, file))
                        ).strftime(TIME_FORMAT),
                        (
                            btyes_to_human_readable(
                                os.path.getsize(os.path.join(path, file))
                            )
                        ),
                    ]
                )
        else:
            # Get defaults or custom selections...
            all_headers = self.header
            assert all_headers is not None, (
                f"`self._header_param_names_default`: Header names are None with parser {self.parser}."
            )
            opt_headers = all_headers[
                len(self.__REQUIRED_HEADER_LIST) :
            ]  # remove 'required' headers.

            print(
                "HERE - adding data with headers: ",
                opt_headers,
                len(self.__REQUIRED_HEADER_LIST),
            )
            # - "Load Success", "#", "Filename", "Created", "Modified", "Size"
            for i, file in enumerate(files):
                # Get the loading index
                load_index = self._status_index()
                data[i] = [i + 1, file]  # add index (1, ...) and filename to data.

                if file in parsers and parsers[file] is not None:
                    # Use existing parser instance, add to files.
                    parser: parserBase = parsers[file]  # Checked above: # type: ignore
                    # Iterate over the optional headers and add them to the data list.
                    for param in opt_headers:
                        if i == 0:
                            print(f"HERE - adding {param}")
                        # Check if the parameter is in the parser.params list
                        if parser is not None and param in parser.params:
                            if param.lower() in [
                                "created",
                                "modified",
                                "ctime",
                                "mtime",
                            ]:
                                # Convert to datetime object
                                value = parser.params[param]
                                if isinstance(value, dt):
                                    data[i].append(value.strftime(TIME_FORMAT))
                                else:
                                    data[i].append(
                                        dt.fromtimestamp(float(value)).strftime(
                                            TIME_FORMAT
                                        )
                                    )

                            elif param.lower() in ["size", "memory_size"]:
                                # Convert to human readable size
                                data[i].append(
                                    btyes_to_human_readable(int(parser.params[param]))
                                )
                            else:
                                # Get the parameter value
                                data[i].append(parser.params[param])
                        # Otherwise check if the parameter is a valid attribute of the parser object.
                        elif parser is not None and hasattr(parser, param):
                            # Get the parameter value
                            data[i].append(getattr(parser, param))
                        else:
                            print(parser, param)
                            # If not, add an empty string to the data list.
                            data[i].append("")

                    status: tableModel.loadStatus
                    if parser.is_loaded:
                        status = tableModel.loadStatus.FULL_DATA
                    else:
                        status = tableModel.loadStatus.HEADER_ONLY
                else:
                    # Not in _parser_headers
                    status = tableModel.loadStatus.ERROR

                data[i].insert(
                    load_index,
                    status,
                    #    (True
                    #    if file in self._parser_headers
                    #    and self._parser_headers[file] is not None
                    #    else False)
                )
                if i == 0:
                    print("The data is ", data[i])
        return data

    def update_header(self) -> None:
        """
        Update the column headers to display.
        """

        header = self.header
        print("The header is ", header)
        if header is None:
            return

        # Format the header
        formatted_header = [label.replace("_", " ").capitalize() for label in header]

        # Ideally just updates header names, but occasionally crashes when attempting to layoutChange.emit()
        if self.files_model is not None:
            selection_model = self.selectionModel()
            if selection_model is not None:
                old_selection = selection_model.selectedRows()
                new_model = tableModel(
                    data=self.files_model._data, header=(formatted_header)
                )
                self.proxy_model.setSourceModel(new_model)
                self.files_model = new_model
                # Setup and restore selection: requires multiselection to select each row before using regular extended selection.
                self.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
                [self.selectRow(i.row()) for i in old_selection]
                self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
            else:
                new_model = tableModel(
                    data=self.files_model._data, header=(formatted_header)
                )
                self.proxy_model.setSourceModel(new_model)
                self.files_model = new_model
                self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        # Update column widths with header update.
        for i in range(0, len(formatted_header)):
            # If the current column width is larger than the size hint, resize to fit for all columns.
            if self.columnWidth(i) > self.sizeHintForColumn(i):
                self.resizeColumnToContents(i)
        return

    def update_table(self):
        """
        Update the directory view.

        Applies text filter, parser and permissionable extensions to filter the files list.
        Loads the filtered header data for the table, appling updates to the class progress bar if defined.
        """
        # Collect the data corresponding to the headers.
        data = self.generate_header_data()

        if len(data) > 1:
            print("The data is ", data[0], data[1], "...")

        # Create a table model
        header = self.header
        self.files_model = tableModel(data=data, header=header)

        # Set the model to the proxy model
        self.proxy_model.setSourceModel(self.files_model)

        # Resize the columns to fit the contents
        self.resizeColumnToContents(0)
        if header is not None:
            for i in range(0, len(header)):
                # If the current column width is larger than the size hint, resize to fit for all columns.
                if self.columnWidth(i) > self.sizeHintForColumn(i):
                    self.resizeColumnToContents(i)
            if self.progress_bar is not None:
                self.progress_bar.setValue(0)

    def load_selection(self) -> None:
        """
        Loads full data from files of the current selection.

        Instead of just header information, loads all data lines
        from the selected files. Loaded datafiles are cached in
        self._parser_files, until the parser object is changed.
        """
        sm = self.selectionModel()
        # Require selection and parser to trigger load data.
        if sm.hasSelection() and self.parser is not None:
            rows = sm.selectedRows(
                column=1 if self._status_index() > 1 else 2
            )  # get filename, not index.

            # Generate a thread pool to load files in parallel.
            pool = QThreadPool.globalInstance()
            """The global thread pool."""
            if pool is None:
                pool = QThreadPool()
            # thread_count: int = pool.maxThreadCount()
            # THREAD_MIN_DIV: int = 5
            # """Minimum number of files to load in a thread."""
            # main_thread_core = QThread.currentThread()
            # """The main thread - leave this core for the GUI."""

            # Set the progress bar to zero.
            if self.progress_bar is not None:
                self.progress_bar.setValue(0)
                # Store number of files to be processed for progress bar
                self.progress_bar.setRange(0, len(rows))

            # Setup warnings to show once.
            warnings.simplefilter("once", UserWarning)

            # Collect files to load in a thread.
            # files_to_load: dict[str, parserBase] = {} # TODO: threading
            for i, row in enumerate(rows):
                filename: str = self.proxy_model.data(row, Qt.ItemDataRole.DisplayRole)
                # Check if the file already loaded:
                if (
                    filename not in self._parser_files
                    or self._parser_files[filename]
                    is None  # re-attempt if parser is None.
                ):
                    # Check for a header object (should always exist).
                    if (
                        filename in self._parser_headers
                        and self._parser_headers[filename] is not None
                    ):
                        # Use existing parser instance, add to files.
                        parser = self._parser_headers[filename]
                        assert isinstance(parser, parserBase)
                        # files_to_load[filename] = parser TODO: threading

                        # TODO: Implement threading generation and signal emit when complete.
                        # if len(files_to_load) > 0:
                        #     if thread_count > 0 and len(files_to_load)//THREAD_MIN_DIV > 0:
                        #         # Setup threads to load files in parallel.
                        #         files_per_thread = ((len(files_to_load) // thread_count)
                        #                             + ((len(files_to_load) % thread_count) > 0)) #Add one to cover remainder if not perfect division
                        #         # Generate threads to load files in parallel.
                        #         threads = [
                        #             fileLoaderWorker(
                        #                 n=i,
                        #                 files=list(files_to_load.keys())[i * files_per_thread : (i + 1) * files_per_thread],
                        #                 parser=self.parser,
                        #                 parsers=files_to_load,
                        #                 progress_bar=self.progress_bar,
                        #                 load_kwargs=dict(header_only = False)
                        #             )
                        #             for i in range(thread_count)
                        #         ]
                        #         # Run the threads
                        #         for thread in threads:
                        #             pool.start(thread)
                        # Wait for the threads to finish
                        #         pool.waitForDone()

                        try:
                            parser.load()  # use internal filepath to load
                            if parser.is_loaded:
                                self._parser_files[filename] = parser
                                ## Update the table icon
                                # Get index from proxy model
                                index = self.proxy_model.index(i, self._status_index())
                                print(
                                    self.proxy_model.data(
                                        index, Qt.ItemDataRole.DecorationRole
                                    ).name()
                                )
                                print(
                                    f"updating to {self.files_model._icon_full_data.name()}, SETDATA"
                                )
                                # self.proxy_model.setData(index,
                                #                          self.files_model._icon_full_data, #type ignore.
                                #                          Qt.ItemDataRole.DecorationRole
                                #                          )
                                assert self.files_model is not None
                                source_index = self.proxy_model.mapToSource(index)
                                self.files_model.setData(
                                    source_index,
                                    self.files_model._icon_full_data,  # type ignore.
                                    Qt.ItemDataRole.DecorationRole,
                                )
                                # Update the model
                                # print(f"Updated index {index}", index.row())
                                print(
                                    self.proxy_model.data(
                                        index, Qt.ItemDataRole.DecorationRole
                                    ).name()
                                )
                                print(
                                    self.files_model.data(
                                        source_index, Qt.ItemDataRole.DecorationRole
                                    ).name()
                                )
                        except Exception as e:
                            traceback.print_exception(e)
                    # else: ignore unloaded headers (None) in the selection.
                if self.progress_bar is not None:
                    self.progress_bar.setValue(i + 1)

            # Restore warnings functionality.
            warnings.resetwarnings()
            # Trigger a loading completed signal.
            self.progress_bar.setValue(0)
            self.selection_loaded.emit(True)
        return

    @property
    def selected_filenames(self) -> list[str]:
        """
        Returns the selected filenames in the directory viewer.

        Returns
        -------
        list[str]
            A list of filenames selected in the directory viewer.
        """
        sm = self.selectionModel()
        if sm.hasSelection():
            rows = sm.selectedRows(
                column=self._index_filename_column
            )  # Filename in second column.
            # return self.files_model.data(rows, Qt.ItemDataRole.DisplayRole)
            return [
                self.proxy_model.data(row, Qt.ItemDataRole.DisplayRole) for row in rows
            ]
        return []

    @property
    def _index_filename_column(self) -> int:
        """
        Returns the index of the filename column in the table model.
        """
        if self.parser is None:
            return 1
        else:
            return 2 if self._status_index() < 2 else 1


class SummaryParamSelector(QTableWidget):
    """
    A togglable interface for selecting parameters of parser objects.
    """

    def __init__(
        self,
        parsers: list[parserBase] | None = None,
        parent: QWidget | None = None,
        existing_selection: list[str] | None = None,
    ):
        # Init the table
        super().__init__(parent)

        # Copy the list of parser_objects
        self._parsers: list[parserBase] = parsers.copy() if parsers is not None else []
        """A list of parser objects to select parameters from."""

        # Create a set of unique parameters
        self._params: set[str] = self._collect_unique_parameters()
        self._filtered_params: set[str] | None = None
        """An optional set of filtered parameters."""
        self.filter_text: str | None = None
        """The current filter text."""

        # Collect the set of parser classes
        self._parser_cls: set[type[parserBase]] = self._collect_unique_parser_cls()
        """A list of the parser subclasses which default `param` values can be taken."""

        # Create a memory set of items (selected or not)
        self._items: dict[str, QTableWidgetItem]
        """A dictionary map of the parameter items that could be in the table, saving their selected state."""
        self._initialise_items()

        # Set a default value for max cols
        self._max_columns: int = 3

        # Disable multiselection
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)

        # Initialise the table view
        self._populate(existing_selection)

    def filter_parameters(self, text: str | None) -> None:
        """
        Filters the parameters by the text string.

        Parameters
        ----------
        text : str | None
            The text to filter the parameters by.
        """
        self.filter_text = text
        if text is None or text == "":
            self._filtered_params = None
            self._populate()
            return
        # Otherwise filter the parameters
        params = self.params
        self._filtered_params = set([p for p in params if text in p])
        self._populate()
        return

    @property
    def max_cols(self):
        """
        The number of columns to spread the parameter display over.
        """
        return self._max_columns

    @max_cols.setter
    def max_cols(self, val: int):
        self._max_columns = val

    @property
    def parsers(self) -> list[parserBase]:
        return self._parsers

    @parsers.setter
    def parsers(self, parsers: list[parserBase] | None) -> None:
        self._parsers = parsers.copy() if parsers is not None else []
        # Update the collection
        self._params = self._collect_unique_parameters()
        # Update the class list
        self._parser_cls = self._collect_unique_parser_cls()
        # Update the table
        self._populate()
        return

    @property
    def params(self) -> set[str]:
        """
        The unique set of parameter names belonging to the list of parsers.
        """
        return self._params

    def _collect_unique_parameters(self) -> set[str]:
        """
        Collects the common parser parameters for each item

        Returns
        -------
        list[str]
            A list of parameter string keys.
        """
        params = set()
        for parser in self.parsers:
            for p in parser.params.keys():
                if p not in params:
                    params.add(p)
        return params

    def _collect_unique_parser_cls(self) -> set[type[parserBase]]:
        """
        Collects the class types in the parser list.

        Returns
        -------
        list[type[parserBase]]
            A list of the unique parser classes in the parser
        """
        cls_list: set[type[parserBase]] = set()
        for p in self._parsers:
            if p.__class__ not in cls_list:
                cls_list.add(p.__class__)
        return cls_list

    def _initialise_items(self) -> None:
        """Initialises the table items with the parameter set and checks the state for summary parameters."""
        self._items = {p: QTableWidgetItem(p) for p in self.params}
        for x in self.params:
            # Make the item uneditable.
            item = self._items[x]
            item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)

            # Check the item if it is a default parameter of any parser classes.
            if any(
                [
                    x in cls.summary_param_names
                    or (
                        x in cls.RELABELS and cls.RELABELS[x] in cls.summary_param_names
                    )
                    for cls in self._parser_cls
                ]
            ):
                # Set the item to checked if it is a default parameter.
                item.setCheckState(Qt.CheckState.Checked)
            else:
                item.setCheckState(Qt.CheckState.Unchecked)
        return

    def _populate(self, existing_selection: list[str] | None = None) -> None:
        """
        Populates the table with the current set of parameter value checkboxes.
        """

        ### TODO: Disconnect the filtering from the active selection returned - currently filtering creates
        # a selection subset of elements you want to tick (and keep track of their tick).

        # Update the existing maps
        if existing_selection:
            for p in existing_selection:
                if p in self._items:
                    self._items[p].setCheckState(Qt.CheckState.Checked)
                else:
                    # Create a new item
                    self._items[p] = QTableWidgetItem(p, Qt.CheckState.Checked.value)

        # Clear the contents, and set the appropriate number of rows/columns
        self.clearContents()
        cols = self.max_cols
        self.setColumnCount(cols)

        # Convert params to an alphabetically ordered list
        if self._filtered_params is not None:
            params = sorted(list(self._filtered_params))
        else:
            params = sorted(list(self.params))

        nparams = len(params)
        # If no parameters, return
        if nparams == 0:
            return
        # Calculate the number of items per column
        items_per_column = nparams // cols
        if items_per_column == 0:
            items_per_column = 1
        # Add extra column for a modulus remainder
        if nparams % items_per_column > 0:
            items_per_column += 1
        self.setRowCount(items_per_column + 1)

        # Additionally, order the default selected parameters first.
        # These need to be checked for existing within the parser param objects, which handle relabelling.
        default_params = []
        for x in params:
            if any(
                [
                    x in cls.summary_param_names
                    or (
                        x in cls.RELABELS and cls.RELABELS[x] in cls.summary_param_names
                    )
                    for cls in self._parser_cls
                ]
            ):
                default_params.append(x)

        params = sorted(params, key=lambda x: x in default_params, reverse=True)

        # Set the items and checked values
        for i, p in enumerate(params):
            # Create the item if it doesn't exist
            if p not in self._items:
                self._items[p] = item = QTableWidgetItem(p)
                # Make the item uneditable.
                item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)
                # Set the check state
                if p in default_params:
                    item.setCheckState(Qt.CheckState.Checked)
                else:
                    item.setCheckState(Qt.CheckState.Unchecked)
            else:
                # Get the item
                item = self._items[p]
            # Assign the item, use a copy so it isn't deleted...
            self.setItem(i % items_per_column, i // items_per_column, item.clone())

        # Adjust the column widths to match min width
        widths = []
        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)
            widths.append(self.columnWidth(i))
        self.setMinimumWidth((sum(widths) * 19) // 18)
        return

    def refresh(self) -> None:
        """
        Updates the table with the current set of parameters.
        """
        self._params = self._collect_unique_parameters()
        self._parser_cls = self._collect_unique_parser_cls()
        self._initialise_items()
        self._populate()
        return

    def resizeEvent(self, e: QtGui.QResizeEvent | None) -> None:
        """
        Resize the columns to fit the new width.
        """
        if e is not None:
            widths = [self.columnWidth(i) for i in range(self.columnCount())]
            scale = self.width() / sum(widths)

            for i in range(self.columnCount()):
                self.setColumnWidth(i, int(widths[i] * scale))
        return super().resizeEvent(e)

    def selection(self) -> set[str]:
        """
        Finds the current item list that has checked values.

        Returns
        -------
        set[str]
            A set of toggled parameter names.
        """
        params = set()
        for i in range(self.rowCount()):
            for j in range(self.columnCount()):
                item = self.item(i, j)
                if item is not None and item.checkState() == Qt.CheckState.Checked:
                    params.add(item.text())
        return params


class SummaryParamSelectorDialog(QtWidgets.QDialog):
    """Dialog for selecting parameters of parser objects."""

    def __init__(
        self,
        parsers: list[parserBase] | None = None,
        parent: QWidget | None = None,
        existing_selection: list[str] | None = None,
    ):
        super().__init__(parent=parent)

        # Setup layout.
        self._layout = QtWidgets.QVBoxLayout()
        self.setLayout(self._layout)

        # Setup margins if a parent is provided to keep interface tight.
        if parent is not None:
            self._layout.setContentsMargins(0, 0, 0, 0)

        # Setup layout and init elements.
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(QtWidgets.QLabel("Select parameters to display:"), stretch=3)
        relabels_checkbox = QtWidgets.QCheckBox("Relabel")
        hlayout.addWidget(relabels_checkbox, stretch=1)
        self._layout.addLayout(hlayout)
        param_filter = QtWidgets.QLineEdit()
        param_filter.setPlaceholderText("Filter parameters...")
        self._layout.addWidget(param_filter)

        # Check if parsers are provided and are relabelled or not. Default all parsers to relabel if not set.
        if parsers is not None and len(parsers) > 0:
            # Check if any parsers are relabelled.
            relabels = [p.relabel for p in parsers]
            # Majority check
            if sum(relabels) > len(relabels) // 2:
                relabels_checkbox.setCheckState(Qt.CheckState.Checked)
                for p in parsers:
                    p.__class__.relabel = True
            else:
                relabels_checkbox.setCheckState(Qt.CheckState.Unchecked)
                for p in parsers:
                    p.__class__.relabel = False

        self._summary_param_selector = SummaryParamSelector(
            parsers=parsers, parent=self, existing_selection=existing_selection
        )
        self._layout.addWidget(self._summary_param_selector)
        self.setWindowTitle("Select Parser Parameters")

        # Add buttons
        QBtn = (
            QtWidgets.QDialogButtonBox.StandardButton.Cancel
            | QtWidgets.QDialogButtonBox.StandardButton.Ok
        )
        self._buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self._buttonBox.accepted.connect(self.accept)
        self._buttonBox.rejected.connect(self.reject)
        self._layout.addWidget(self._buttonBox)

        # Bind the relabel checkbox to the parser objects.
        relabels_checkbox.stateChanged.connect(self._relabel_parser_objects)

        # Bind the filter line edit to the parser objects.
        param_filter.textChanged.connect(self._filter_parameters)

    def _filter_parameters(self, text: str) -> None:
        """
        Filters the parameters by the text string.

        Parameters
        ----------
        text : str | None
            The text to filter the parameters by.
        """
        self.selector.filter_parameters(text)
        return

    def _relabel_parser_objects(self, state: int) -> None:
        """
        Relabels the parser objects if the checkbox is toggled.

        Parameters
        ----------
        state : int
            The state of the checkbox
        """
        # Convert the int to the enumerate.
        state = Qt.CheckState(state)
        # Check the state of the checkbox
        for cls in self.selector._parser_cls:
            cls.relabel = state is Qt.CheckState.Checked
        # Repopulate the table with (re)labelled parameters.
        self.selector.refresh()
        return

    @property
    def selector(self) -> SummaryParamSelector:
        """
        The SummaryParamSelector object.

        Returns
        -------
        SummaryParamSelector
            The SummaryParamSelector object.
        """
        return self._summary_param_selector

    @property
    def selection(self) -> set[str]:
        """
        The SummaryParamSelector toggled selection.

        Returns
        -------
        set[str]
            A set of selected parameter names.
        """
        return self.selector.selection()

    @override
    def keyPressEvent(self, a0: QtGui.QKeyEvent | None) -> None:
        if a0 is None:
            return
        if a0.key() == QtCore.Qt.Key.Key_Escape:
            self.reject()
        elif (
            a0.key() == QtCore.Qt.Key.Key_Return or a0.key() == QtCore.Qt.Key.Key_Enter
        ):
            return a0.accept()
        else:
            super().keyPressEvent(a0)

    @override
    def accept(self):
        print("OK")
        return super().accept()

    @override
    def reject(self):
        return super().reject()


# Example
if __name__ == "__main__":
    # Load some example parsers
    base_dir = os.getcwd()
    f1 = r"tests\\test_data\\au\\MEX2\\2024-03\\MEX2_5640.mda"
    f2 = r"tests\\test_data\\au\\MEX2\\2024-03\\MEX2_5642.mda"
    files = [os.path.normpath(os.path.join(base_dir, f)) for f in [f1, f2]]

    # Relabelling
    MEX2_NEXAFS.relabel = True
    parsers = [MEX2_NEXAFS(f, header_only=True) for f in files]

    app = QApplication(sys.argv)
    window = SummaryParamSelectorDialog(parsers=parsers)
    # window = directoryViewerTableNew(os.path.dirname(files[0]))
    window.show()
    window.setWindowTitle("Summary Param Selector")
    val = app.exec()
    print(val)
    if val:
        print(window.selection)
        sys.exit(True)
    else:
        sys.exit(False)
